/*
 * Copyright (c) 2010 Yoshikazu Kuramochi
 * All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package ch.kuramo.javie.app;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Arrays;
import java.util.Comparator;
import java.util.regex.Pattern;

public class IntelGMAWorkaround {

	public static void workaround() {
		if (!System.getProperty("os.name").toLowerCase().contains("windows")) {
			return;
		}

		File workingDir = new File(System.getProperty("user.dir"));
		File logFile = lastJvmLogFile(workingDir);
		if (logFile == null) return;

		try {
			if (!gmaCrashed(logFile)) return;

			System.setProperty("sun.java2d.opengl", "false");
			createNewIniFile(workingDir);

		} catch (IOException e) {
			e.printStackTrace();
			// ignore
		}
	}

	private static File lastJvmLogFile(File workingDir) {
		// JVMが吐き出したログファイルをリストアップする。
		final Pattern namePat = Pattern.compile("^hs_err_pid\\d+\\.log$");
		File[] logFiles = workingDir.listFiles(new FileFilter() {
			public boolean accept(File pathname) {
				return pathname.isFile() && namePat.matcher(pathname.getName()).matches();
			}
		});
		if (logFiles.length == 0) {
			return null;
		}

		// 新しい順にソート。
		Arrays.sort(logFiles, new Comparator<File>() {
			public int compare(File o1, File o2) {
				return (int)Math.signum(o2.lastModified() - o1.lastModified());
			}
		});

		// 一番新しいログファイルが１時間以上経過している場合は無視。
		File lastLogFile = logFiles[0];
		if (System.currentTimeMillis() - lastLogFile.lastModified() >= 3600000) {
			return null;
		}

		return lastLogFile;
	}

	private static boolean gmaCrashed(File logFile) throws IOException {
		// GMAのドライバ内でクラッシュしているかどうかを調べる。
		BufferedReader br = null;
		try {
			br = new BufferedReader(new FileReader(logFile));
			String line;
			while ((line = br.readLine()) != null) {
				if (line.startsWith("C  [ig4icd32.dll+") || line.contains("C  [ig4icd64.dll+")) {
					return true;
				}
			}
			return false;
		} finally {
			if (br != null) br.close();
		}
	}

	private static void createNewIniFile(File dir) throws IOException {
		File iniFile = new File(dir, "Javie.ini");

		String newIni = createNewIni(iniFile);
		if (newIni == null) return;

		File bakFile = new File(dir, "Javie.ini.bak");
		for (int i = 1; bakFile.exists(); ++i) {
			bakFile = new File(dir, "Javie.ini.bak-"+i);
		}
		iniFile.renameTo(bakFile);

		FileWriter fw = new FileWriter(iniFile);
		try {
			fw.write(newIni);
		} finally {
			fw.close();
		}
	}

	private static String createNewIni(File iniFile) throws IOException {
		boolean create = false;
		StringWriter sw = new StringWriter((int)iniFile.length());

		BufferedReader br = null;
		PrintWriter pw = null;
		try {
			br = new BufferedReader(new FileReader(iniFile));
			pw = new PrintWriter(sw);
			String line;
			while ((line = br.readLine()) != null) {
				if (line.equals("-Dsun.java2d.opengl=true")) {
					line = "-Dsun.java2d.opengl=false";
					create = true;
				}
				pw.println(line);
			}
		} finally {
			if (pw != null) pw.close();
			if (br != null) br.close();
		}

		return create ? sw.toString() : null;
	}


	private IntelGMAWorkaround() { }

}
