/*
 * Copyright (c) 2009,2010 Yoshikazu Kuramochi
 * All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package ch.kuramo.javie.core.shaders;

import java.util.Arrays;
import java.util.List;

import javax.media.opengl.GL2;
import javax.media.opengl.GLContext;
import javax.media.opengl.Threading;

import ch.kuramo.javie.api.annotations.GLProgram;
import ch.kuramo.javie.api.annotations.GLShader;
import ch.kuramo.javie.api.annotations.GLShader.ShaderType;
import ch.kuramo.javie.core.Util;
import ch.kuramo.javie.core.services.GLGlobal;

import com.google.inject.Inject;

public class BlendModeShaders {

	private static final String[] createProgramSource(String name, boolean temporal) {
		return new String[] {
								"uniform sampler2DRect texDst;",
								"uniform sampler2DRect texSrc;",
								"uniform float opacity;",
				temporal ?		"uniform float time;" : "",
								"",
				String.format(	"vec4 blend_%s(vec4 pDst, vec4 pSrc, float opacity%s);", name, temporal ? ", float time" : ""),
								"",
								"void main(void)",
								"{",
								"	vec2 texCoord = gl_TexCoord[0].st;",
								"	vec4 dst = texture2DRect(texDst, texCoord);",
								"	vec4 src = texture2DRect(texSrc, texCoord);",
				String.format(	"	dst = blend_%s(vec4(dst.rgb*dst.a, dst.a), src, opacity%s);", name, temporal ? ", time" : ""),
								"	gl_FragColor = (dst.a != 0.0) ? vec4(dst.rgb/dst.a, dst.a) : vec4(0.0);",
								"}"
		};
	}

	private static final String[] createProgramSource(String name) {
		return createProgramSource(name, false);
	}


	@GLShader(ShaderType.FRAGMENT_SHADER)
	public static final String blend_functions = "blend_functions.frag";


	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] NORMAL = createProgramSource("normal");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach={"blend_functions"})
	public static final String[] DISSOLVE = createProgramSource("dissolve");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach={"blend_functions"})
	public static final String[] DANCING_DISSOLVE = createProgramSource("dancing_dissolve", true);

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] DARKEN = createProgramSource("darken");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] MULTIPLY = createProgramSource("multiply");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] COLOR_BURN = createProgramSource("color_burn");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] LINEAR_BURN = createProgramSource("linear_burn");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] DARKER_COLOR = createProgramSource("darker_color");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] ADD = createProgramSource("add");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] LIGHTEN = createProgramSource("lighten");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] SCREEN = createProgramSource("screen");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] COLOR_DODGE = createProgramSource("color_dodge");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] LINEAR_DODGE = createProgramSource("linear_dodge");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] LIGHTER_COLOR = createProgramSource("lighter_color");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] OVERLAY = createProgramSource("overlay");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] SOFT_LIGHT = createProgramSource("soft_light");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] HARD_LIGHT = createProgramSource("hard_light");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] LINEAR_LIGHT = createProgramSource("linear_light");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] VIVID_LIGHT = createProgramSource("vivid_light");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] PIN_LIGHT = createProgramSource("pin_light");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] HARD_MIX = createProgramSource("hard_mix");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] DIFFERENCE = createProgramSource("difference");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] EXCLUSION = createProgramSource("exclusion");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] HUE = createProgramSource("hue");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] SATURATION = createProgramSource("saturation");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] COLOR = createProgramSource("color");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] LUMINOSITY = createProgramSource("luminosity");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] STENCIL_ALPHA = createProgramSource("stencil_alpha");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] STENCIL_LUMA = createProgramSource("stencil_luma");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] SILHOUETTE_ALPHA = createProgramSource("silhouette_alpha");

	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public static final String[] SILHOUETTE_LUMA = createProgramSource("silhouette_luma");



	@GLShader(ShaderType.FRAGMENT_SHADER)
	@GLProgram(attach="blend_functions")
	public final String[] INTERSECTING_BLEND;


	private static int _maxIntersectingLayers;

	public static int getMaxIntersectingLayers() {
		return _maxIntersectingLayers;
	}


	private String[] createIntersectingBlendSource(boolean intel) {
		List<String> source = Util.newList();

		// 以下のような変な書き方をしているのは、普通に書いたらMacで動かなかったから。
		// (sampler2DRectの配列をuniform変数にしても、配列のインデックスを定数にしないと動かない)

		// あと、シェーダ内で配列を使うと遅くなるようなので、ord0-ordNは配列にしていない。
		// TODO texture2DRectでサンプリングした値は変数に保持しておいた方がいいかも。
		//      ただし配列ではなくord0-ordNと同じような形で。


		/********************
		const int maxLayers = N;

		uniform int nLayers;
		uniform sampler2DRect texDst;
		uniform sampler2DRect texSrc0;
		uniform sampler2DRect texDep0;
		...
		uniform sampler2DRect texSrcN;
		uniform sampler2DRect texDepN;
		uniform int blendMode[maxLayers];
		uniform float opacity[maxLayers];
		uniform float time;

		vec2 texCoord = gl_TexCoord[0].st;
		********************/

		source.add(String.format("const int maxLayers = %d;", _maxIntersectingLayers));

		source.add("uniform int nLayers;");
		source.add("uniform sampler2DRect texDst;");

		for (int i = 0; i < _maxIntersectingLayers; ++i) {
			source.add(String.format("uniform sampler2DRect texSrc%d;", i));
			source.add(String.format("uniform sampler2DRect texDep%d;", i));
		}

		if (!intel) source.add("uniform int blendMode[maxLayers];");

		source.add("uniform float opacity[maxLayers];");
		source.add("uniform float time;");

		source.add("vec2 texCoord = gl_TexCoord[0].st;");


		/********************
		vec4 getSrc(int i) {
			return
				(i == 0) ? texture2DRect(texSrc0, texCoord) :
				...
				(i == N) ? texture2DRect(texSrcN, texCoord) :
				vec4(0.0);
		}
		********************/

		source.add("vec4 getSrc(int i) {");
		source.add("	return");
		for (int i = 0; i < _maxIntersectingLayers; ++i) {
			source.add(String.format("		(i == %1$d) ? texture2DRect(texSrc%1$d, texCoord) :", i));
		}
		source.add("		vec4(0.0);");
		source.add("}");


		/********************
		float getDep(int i) {
			return
				(i == 0) ? texture2DRect(texDep0, texCoord).r :
				...
				(i == N) ? texture2DRect(texDepN, texCoord).r :
				-1.0;
		}
		********************/

		source.add("float getDep(int i) {");
		source.add("	return");
		for (int i = 0; i < _maxIntersectingLayers; ++i) {
			source.add(String.format("		(i == %1$d) ? texture2DRect(texDep%1$d, texCoord).r :", i));
		}
		source.add("		-1.0;");
		source.add("}");


		/********************
		int ord0 = 0;
		...
		int ordN = N;
		********************/

		for (int i = 0; i < _maxIntersectingLayers; ++i) {
			source.add(String.format("int ord%1$d = %1$d;", i));
		}


		/********************
		int getOrd(int i) {
			return
				(i == 0) ? ord0 :
				...
				(i == N) ? ordN :
				-1;
		}
		********************/

		source.add("int getOrd(int i) {");
		source.add("	return");
		for (int i = 0; i < _maxIntersectingLayers; ++i) {
			source.add(String.format("		(i == %1$d) ? ord%1$d :", i));
		}
		source.add("		-1;");
		source.add("}");


		/********************
		void setOrd(int i, int ord) {
			if (i == 0) ord0 = ord; else
			...
			if (i == N) ordN = ord;
		}
		********************/

		source.add("void setOrd(int i, int ord) {");
		for (int i = 0; i < _maxIntersectingLayers-1; ++i) {
			source.add(String.format("	if (i == %1$d) ord%1$d = ord; else ", i));
		}
		source.add(String.format("	if (i == %1$d) ord%1$d = ord;", _maxIntersectingLayers-1));
		source.add("}");


		/********************/

		source.addAll(Arrays.asList(new String[] {
	   (intel ?	"vec4 blend_normal(vec4 pDst, vec4 pSrc, float opacity);"
			  :	"vec4 blend(int mode, vec4 pDst, vec4 pSrc, float opacity, float time);"),
				"",
				"void main(void)",
				"{",
				"	vec4 dst = texture2DRect(texDst, texCoord);",
				"",
				// texSrcNのアルファ値が全てゼロ(nzaCount == 0)の場合、texDstの値をそのまま使えばよい。
				// texSrcNのうちひとつだけがゼロ以外のアルファを持つ(nzaCount == 1)場合、そのテクスチャだけをtexDstとブレンドすればよい。
				"	int nzaCount = 0;",
				"	int nzaLayer;",
				"	vec4 nzaColor;",
				"	for (int i = 0; i < nLayers; ++i) {",
				"		vec4 s = getSrc(i);",
				"		if (s.a > 0.0) {",
				"			if (++nzaCount > 1) {",
				"				break;",
				"			}",
				"			nzaLayer = i;",
				"			nzaColor = s;",
				"		}",
				"	}",
				"	if (nzaCount == 1) {",
				"		dst = vec4(dst.rgb*dst.a, dst.a);",
	   (intel ?	"		dst = blend_normal(dst, nzaColor, opacity[nzaLayer]);"
			  :	"		dst = blend(blendMode[nzaLayer], dst, nzaColor, opacity[nzaLayer], time);"),
				"		dst = (dst.a != 0.0) ? vec4(dst.rgb/dst.a, dst.a) : vec4(0.0);",
				"",
				"	} else if (nzaCount > 1) {",
				"		dst = vec4(dst.rgb*dst.a, dst.a);",
				"		",
				"		for (int i = nLayers - 2; i >= 0; --i) {",
				"			int ord = getOrd(i);",
				"			float dep = getDep(ord);",
				"			int j, ord2;",
				"			for (j = i + 1; j < nLayers && getDep(ord2 = getOrd(j)) > dep; ++j) {",
				"				setOrd(j - 1, ord2);",
				"			}",
				"			setOrd(j - 1, ord);",
				"		}",
				"		",
				"		for (int i = 0; i < nLayers; ++i) {",
				"			int ord = getOrd(i);",
	   (intel ?	"			dst = blend_normal(dst, getSrc(ord), opacity[ord]);"
			  :	"			dst = blend(blendMode[ord], dst, getSrc(ord), opacity[ord], time);"),
				"		}",
				"		dst = (dst.a != 0.0) ? vec4(dst.rgb/dst.a, dst.a) : vec4(0.0);",
				"	}",
				"",
				"	gl_FragColor = dst;",
				"}"
		}));

		return source.toArray(new String[source.size()]);
	}

	@Inject
	public BlendModeShaders(final GLGlobal glGlobal) {
		final int[] texUnits = new int[1];
		Threading.invokeOnOpenGLThread(new Runnable() {
			public void run() {
				GLContext context = glGlobal.getSharedContext();
				context.makeCurrent();
				try {
					GL2 gl = context.getGL().getGL2();
					gl.glGetIntegerv(GL2.GL_MAX_TEXTURE_IMAGE_UNITS, texUnits, 0);
				} finally {
					context.release();
				}
			}
		});

		// TODO 32個より多くのテクスチャユニットを扱えるのかどうか？
		//
		// GL_TEXTURE* は GL_TEXTURE0 から GL_TEXTURE31 の32個しか定義されていない。
		// OpenGLの仕様上、テクスチャユニットは32個までということなのだろうか？
		// とりあえず32個が上限と仮定して制限しておく。
		// （あまり多く交差するのはパフォーマンス的な問題もあるし）
		//
		_maxIntersectingLayers = (Math.min(32, texUnits[0]) - 1) / 2;

		INTERSECTING_BLEND = createIntersectingBlendSource(glGlobal.isIntel());
	}

}
